// Package v1 contains API Schema definitions for the lifecycle v1 API group
// +groupName=lifecycle.keptn.sh
// +versionName=v1
package common

import (
	"crypto/sha256"
	"encoding/hex"
	"math/rand"
	"strconv"

	operatorcommon "github.com/keptn/lifecycle-toolkit/lifecycle-operator/common"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/metric"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

const WorkloadAnnotation = "keptn.sh/workload"
const VersionAnnotation = "keptn.sh/version"
const AppAnnotation = "keptn.sh/app"
const PreDeploymentTaskAnnotation = "keptn.sh/pre-deployment-tasks"
const PostDeploymentTaskAnnotation = "keptn.sh/post-deployment-tasks"
const K8sRecommendedWorkloadAnnotations = "app.kubernetes.io/name"
const K8sRecommendedVersionAnnotations = "app.kubernetes.io/version"
const K8sRecommendedAppAnnotations = "app.kubernetes.io/part-of"
const K8sRecommendedManagedByAnnotations = "app.kubernetes.io/managed-by"
const PreDeploymentEvaluationAnnotation = "keptn.sh/pre-deployment-evaluations"
const PostDeploymentEvaluationAnnotation = "keptn.sh/post-deployment-evaluations"
const SchedulingGateRemoved = "keptn.sh/scheduling-gate-removed"
const TaskNameAnnotation = "keptn.sh/task-name"
const NamespaceEnabledAnnotation = "keptn.sh/lifecycle-toolkit"
const CreateAppTaskSpanName = "create_%s_app_task"
const CreateWorkloadTaskSpanName = "create_%s_deployment_task"
const CreateAppEvalSpanName = "create_%s_app_evaluation"
const CreateWorkloadEvalSpanName = "create_%s_deployment_evaluation"
const AppTypeAnnotation = "keptn.sh/app-type"
const KeptnGate = "keptn-prechecks-gate"
const ContainerNameAnnotation = "keptn.sh/container"
const MetadataAnnotation = "keptn.sh/metadata"

const MinKeptnNameLen = 80
const MaxK8sObjectLength = 253

type AppType string

const (
	AppTypeSingleService AppType = "single-service"
	AppTypeMultiService  AppType = "multi-service"
)

// KeptnState  is a string containing current Phase state  (Progressing/Succeeded/Failed/Unknown/Pending/Deprecated/Warning)
type KeptnState string

const (
	StateProgressing KeptnState = "Progressing"
	StateSucceeded   KeptnState = "Succeeded"
	StateFailed      KeptnState = "Failed"
	StateUnknown     KeptnState = "Unknown"
	StatePending     KeptnState = "Pending"
	StateDeprecated  KeptnState = "Deprecated"
	StateWarning     KeptnState = "Warning"
)

func (k KeptnState) IsCompleted() bool {
	return k == StateSucceeded || k == StateFailed || k == StateDeprecated || k == StateWarning
}

func (k KeptnState) IsSucceeded() bool {
	return k == StateSucceeded
}

func (k KeptnState) IsFailed() bool {
	return k == StateFailed
}

func (k KeptnState) IsDeprecated() bool {
	return k == StateDeprecated
}

func (k KeptnState) IsPending() bool {
	return k == StatePending
}

func (k KeptnState) IsWarning() bool {
	return k == StateWarning
}

type StatusSummary struct {
	Total       int
	Progressing int
	Failed      int
	Succeeded   int
	Pending     int
	Unknown     int
	Deprecated  int
}

func UpdateStatusSummary(status KeptnState, summary StatusSummary) StatusSummary {
	switch status {
	case StateFailed:
		summary.Failed++
	case StateDeprecated:
		summary.Deprecated++
	case StateSucceeded:
		summary.Succeeded++
	case StateProgressing:
		summary.Progressing++
	case StatePending, "":
		summary.Pending++
	case StateUnknown:
		summary.Unknown++
	}
	return summary
}

func (s StatusSummary) GetTotalCount() int {
	return s.Failed + s.Succeeded + s.Progressing + s.Pending + s.Unknown + s.Deprecated
}

func GetOverallState(s StatusSummary) KeptnState {
	if s.Failed > 0 || s.Deprecated > 0 {
		return StateFailed
	}
	if s.Progressing > 0 {
		return StateProgressing
	}
	if s.Pending > 0 {
		return StatePending
	}
	if s.Unknown > 0 {
		return StateUnknown
	}
	if s.GetTotalCount() != s.Total {
		return StatePending
	}
	return StateSucceeded
}

func GetOverallStateBlockedDeployment(s StatusSummary, blockedDeployment bool) KeptnState {
	state := GetOverallState(s)
	if !blockedDeployment && state == StateFailed {
		return StateWarning
	}
	return state
}

func TruncateString(s string, max int) string {
	if len(s) > max {
		return s[:max]
	}
	return s
}

func Hash(num int64) string {
	// generate the SHA-256 hash of the bytes
	hash := sha256.Sum256([]byte(strconv.FormatInt(num, 10)))
	// take the first 4 bytes of the hash and convert to hex
	return hex.EncodeToString(hash[:4])
}

type CheckType string

const PreDeploymentCheckType CheckType = "pre"
const PostDeploymentCheckType CheckType = "post"
const PromotionCheckType CheckType = "promotion"
const PreDeploymentEvaluationCheckType CheckType = "pre-eval"
const PostDeploymentEvaluationCheckType CheckType = "post-eval"

type KeptnMeters struct {
	TaskCount          metric.Int64Counter
	TaskDuration       metric.Float64Histogram
	DeploymentCount    metric.Int64Counter
	DeploymentDuration metric.Float64Histogram
	AppCount           metric.Int64Counter
	AppDuration        metric.Float64Histogram
	EvaluationCount    metric.Int64Counter
	EvaluationDuration metric.Float64Histogram
	PromotionCount     metric.Int64Counter
}

const (
	AppName                 attribute.Key = attribute.Key("keptn.deployment.app.name")
	AppVersion              attribute.Key = attribute.Key("keptn.deployment.app.version")
	AppNamespace            attribute.Key = attribute.Key("keptn.deployment.app.namespace")
	AppStatus               attribute.Key = attribute.Key("keptn.deployment.app.status")
	AppPreviousVersion      attribute.Key = attribute.Key("keptn.deployment.app.previousversion")
	WorkloadName            attribute.Key = attribute.Key("keptn.deployment.workload.name")
	WorkloadVersion         attribute.Key = attribute.Key("keptn.deployment.workload.version")
	WorkloadPreviousVersion attribute.Key = attribute.Key("keptn.deployment.workload.previousversion")
	WorkloadNamespace       attribute.Key = attribute.Key("keptn.deployment.workload.namespace")
	WorkloadStatus          attribute.Key = attribute.Key("keptn.deployment.workload.status")
	TaskStatus              attribute.Key = attribute.Key("keptn.deployment.task.status")
	TaskName                attribute.Key = attribute.Key("keptn.deployment.task.name")
	TaskType                attribute.Key = attribute.Key("keptn.deployment.task.type")
	EvaluationStatus        attribute.Key = attribute.Key("keptn.deployment.evaluation.status")
	EvaluationName          attribute.Key = attribute.Key("keptn.deployment.evaluation.name")
	EvaluationType          attribute.Key = attribute.Key("keptn.deployment.evaluation.type")
)

func GenerateTaskName(checkType CheckType, taskName string) string {
	randomId := rand.Intn(99_999-10_000) + 10000
	return operatorcommon.CreateResourceName(MaxK8sObjectLength, MinKeptnNameLen, string(checkType), taskName, strconv.Itoa(randomId))
}

func GenerateJobName(taskName string) string {
	randomId := rand.Intn(99_999-10_000) + 10000
	return operatorcommon.CreateResourceName(MaxK8sObjectLength, MinKeptnNameLen, taskName, strconv.Itoa(randomId))
}

func GenerateEvaluationName(checkType CheckType, evalName string) string {
	randomId := rand.Intn(99_999-10_000) + 10000
	return operatorcommon.CreateResourceName(MaxK8sObjectLength, MinKeptnNameLen, string(checkType), evalName, strconv.Itoa(randomId))
}

// MergeMaps merges two maps into a new map. If a key exists in both maps, the
// value of the second map is picked.
func MergeMaps(m1 map[string]string, m2 map[string]string) map[string]string {
	merged := make(map[string]string, len(m1)+len(m2))
	for key, value := range m1 {
		merged[key] = value
	}
	for key, value := range m2 {
		merged[key] = value
	}
	return merged
}

// IsOwnerSupported returns whether the owner of the given object is supported to be considered a KeptnWorkload
func IsOwnerSupported(owner metav1.OwnerReference) bool {
	return owner.Kind == "ReplicaSet" || owner.Kind == "Deployment" || owner.Kind == "StatefulSet" || owner.Kind == "DaemonSet" || owner.Kind == "Rollout"
}
